"""Module that contains dotenv parsers."""

from __future__ import annotations

import re
import codecs
import dataclasses

from typing import IO
from typing import Iterator
from typing import Match
from typing import NamedTuple
from typing import Optional
from typing import Pattern
from typing import Tuple


Original = NamedTuple(
    "Original",
    [
        ("string", str),
        ("line", int),
    ],
)

Binding = NamedTuple(
    "Binding",
    [
        ("key", Optional[str]),
        ("value", Optional[str]),
        ("original", Original),
        ("error", bool),
    ],
)


class DotenvParseError(Exception):
    """Exception for dotenv parse errors."""


def make_regex(string: str, extra_flags: int = 0) -> Pattern[str]:
    """Construct a regex expression."""

    return re.compile(string, re.UNICODE | extra_flags)


_newline = make_regex(r"(\r\n|\n|\r)")
_multiline_whitespace = make_regex(r"\s*", extra_flags=re.MULTILINE)
_whitespace = make_regex(r"[^\S\r\n]*")
_export = make_regex(r"(?:export[^\S\r\n]+)?")
_single_quoted_key = make_regex("'([^']+)'")
_unquoted_key = make_regex(r"([^=\#\s]+)")
_equal_sign = make_regex(r"(=[^\S\r\n]*)")
_single_quoted_value = make_regex(r"'((?:\\'|[^'])*)'")
_double_quoted_value = make_regex(r'"((?:\\"|[^"])*)"')
_unquoted_value = make_regex(r"([^\r\n]*)")
_comment = make_regex(r"(?:[^\S\r\n]*#[^\r\n]*)?")
_end_of_line = make_regex(r"[^\S\r\n]*(?:\r\n|\n|\r|$)")
_rest_of_line = make_regex(r"[^\r\n]*(?:\r|\n|\r\n)?")
_double_quote_escapes = make_regex(r"\\[\\'\"abfnrtv]")
_single_quote_escapes = make_regex(r"\\[\\']")


@dataclasses.dataclass
class Position(object):
    """Model of a cursor position."""

    chars: int
    line: int

    @classmethod
    def start(cls) -> Position:
        """Get a start position."""

        return cls(chars=0, line=1)

    def set(self, other: Position) -> None:
        """Set a position."""

        self.chars = other.chars
        self.line = other.line

    def advance(self, string: str) -> None:
        """Update a position."""

        self.chars += len(string)
        self.line += len(re.findall(_newline, string))


class Reader(object):
    """Dotenv reader."""

    def __init__(self, stream: IO[str]) -> None:
        """Initialize."""

        self.string = stream.read()
        self.position = Position.start()
        self.mark = Position.start()

    def has_next(self) -> bool:
        """Check if a dotenv has the next position."""

        return self.position.chars < len(self.string)

    def set_mark(self) -> None:  # noqa: WPS615
        """Set a mark."""

        # WPS615 - this setter is not a standard OOP setter

        self.mark.set(self.position)

    def get_marked(self) -> Original:
        """Get a mark."""

        # fmt: off
        return Original(
            string=self.string[self.mark.chars: self.position.chars],
            line=self.mark.line,
        )
        # fmt: on

    def peek(self, count: int) -> str:
        """Peek a dotenv."""

        # fmt: off
        return self.string[self.position.chars: self.position.chars + count]
        # fmt: on

    def read_regex(self, regex: Pattern[str]) -> Tuple[str, ...]:
        """Read a dotenv with a regex."""

        match = regex.match(self.string, self.position.chars)

        if match is None:
            raise DotenvParseError("Pattern not found.")

        # fmt: off
        matched = self.string[match.start(): match.end()]
        # fmt: on

        self.position.advance(matched)
        return match.groups()


def decode_match(match: Match[str]) -> str:
    """Decode matches."""

    return codecs.decode(match.group(0), "unicode-escape")


def decode_escapes(regex: Pattern[str], string: str) -> str:
    """Decode escape symbols."""

    return regex.sub(decode_match, string)


def parse_key(reader: Reader) -> Optional[str]:
    """Parse a dotenv key."""

    char = reader.peek(1)

    if char == "#":
        return None

    elif char == "'":
        key, *_ = reader.read_regex(_single_quoted_key)

    else:
        key, *_ = reader.read_regex(_unquoted_key)

    return key


def parse_unquoted_value(reader: Reader) -> str:
    """Parse an unquoted dotenv value."""

    part, *_ = reader.read_regex(_unquoted_value)
    return re.sub(r"\s+#.*", "", part).rstrip()


def parse_value(reader: Reader) -> str:
    """Parse a dotenv value."""

    char = reader.peek(1)

    if char == "'":
        value, *_ = reader.read_regex(_single_quoted_value)
        return decode_escapes(_single_quote_escapes, value)

    elif char == '"':
        value, *_ = reader.read_regex(_double_quoted_value)
        return decode_escapes(_double_quote_escapes, value)

    elif char in {"", "\n", "\r"}:
        return ""

    return parse_unquoted_value(reader)


def parse_binding(reader: Reader) -> Binding:
    """Parse a dotenv binding."""

    reader.set_mark()

    # WPS229 - this part is a part of the original package
    try:  # noqa: WPS229
        reader.read_regex(_multiline_whitespace)

        if not reader.has_next():
            return Binding(
                key=None,
                value=None,
                original=reader.get_marked(),
                error=False,
            )

        reader.read_regex(_export)
        key = parse_key(reader)
        reader.read_regex(_whitespace)

        if reader.peek(1) == "=":
            reader.read_regex(_equal_sign)
            value = parse_value(reader)

        else:
            value = None

        reader.read_regex(_comment)
        reader.read_regex(_end_of_line)

        return Binding(
            key=key,
            value=value,
            original=reader.get_marked(),
            error=False,
        )

    except DotenvParseError:
        reader.read_regex(_rest_of_line)
        return Binding(
            key=None,
            value=None,
            original=reader.get_marked(),
            error=True,
        )


def parse_stream(stream: IO[str]) -> Iterator[Binding]:
    """Parse a dotenv stream."""

    reader = Reader(stream)

    while reader.has_next():
        yield parse_binding(reader)



"""poetry-dotenv - is the plugin that automatically loads environment variables from a dotenv file into the environment before poetry commands are run."""

__title__ = "poetry-dotenv"
__summary__ = "poetry-dotenv - is the plugin that automatically loads environment variables from a dotenv file into the environment before poetry commands are run."
__uri__ = "https://github.com/volopivoshenko/poetry-dotenv"

__version__ = "0.2.2"

__author__ = "Volodymyr Pivoshenko"
__email__ = "volodymyr.pivoshenko@gmail.com"

__license__ = "MIT"
__copyright__ = "Copyright 2022, Volodymyr Pivoshenko"

from poetry_dotenv import dotenv
from poetry_dotenv import plugin



"""Module that contains the core functionality of the plugin."""

import os

from cleo.events.console_command_event import ConsoleCommandEvent
from cleo.events.console_events import COMMAND
from poetry.console.application import Application
from poetry.console.commands.env_command import EnvCommand
from poetry.plugins.application_plugin import ApplicationPlugin

from poetry_dotenv import dotenv


class DotenvPlugin(ApplicationPlugin):
    """Plugin that automatically loads environment variables from a dotenv file.

    Plugin that automatically loads environment variables from a dotenv file into the environment
    before ``poetry`` commands are run.

    Notes
    -----
    To prevent ``poetry`` from loading the dotenv file, set the ``POETRY_DONT_LOAD_ENV``
    environment variable.

    If your dotenv file is located in a different path or has a different name you may set
    the ``POETRY_DOTENV_LOCATION`` environment variable.
    """

    debug_msg = "<debug>{0!s}</debug>"

    def activate(self, application: Application) -> None:
        """Activate the plugin."""

        application.event_dispatcher.add_listener(  # pragma: no cover
            event_name=COMMAND,
            listener=self.load,
        )

    def debug(self, event: ConsoleCommandEvent, message: str) -> None:  # pragma: no cover
        """Log a debug message."""

        if event.io.is_debug():
            event.io.write_line(self.debug_msg.format(message))

    def load(self, event: ConsoleCommandEvent, *args, **kwargs) -> None:
        """Load a dotenv file."""

        dont_load_dotenv = os.getenv("POETRY_DONT_LOAD_DOTENV")
        dotenv_location = os.getenv("POETRY_DOTENV_LOCATION")
        is_env_command = isinstance(event.command, EnvCommand)

        if is_env_command and dont_load_dotenv:
            self.debug(event, "Not loading environment variables.")

        elif is_env_command and not dont_load_dotenv:
            filepath = dotenv_location if dotenv_location else dotenv.core.find(usecwd=True)

            if os.path.isfile(filepath):
                self.debug(event, "Loading environment variables from {0!r}.".format(filepath))
                dotenv.core.load(filepath=filepath)

            else:
                self.debug(event, "File {0!r} doesn't exist.".format(filepath))



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['poetry_dotenv', 'poetry_dotenv.dotenv']

package_data = \
{'': ['*']}

install_requires = \
['poetry>=1.2.2,<2.0.0']

entry_points = \
{'poetry.application.plugin': ['poetry-dotenv = '
                               'poetry_dotenv.plugin:DotenvPlugin']}

setup_kwargs = {
    'name': 'poetry-dotenv',
    'version': '0.2.2',
    'description': 'poetry-dotenv - is the plugin that automatically loads environment variables from a dotenv file into the environment before poetry commands are run.',
    'long_description': '<div align="center">\n  <a href="https://pypi.org/project/poetry-dotenv">\n    <img alt="logo" src="https://github.com/volopivoshenko/poetry-dotenv/blob/main/docs/static/assets/logo.svg?raw=True" height=200>\n  </a>\n</div>\n\n<p align="center">\n  <a href="https://opensource.org/licenses/MIT">\n    <img alt="license" src="https://img.shields.io/pypi/l/poetry-dotenv?logo=opensourceinitiative">\n  </a>\n  <a href="https://pypi.org/project/poetry-dotenv">\n    <img alt="python" src="https://img.shields.io/pypi/pyversions/poetry-dotenv?logo=python">\n  </a>\n  <a href="https://pypi.org/project/poetry-dotenv">\n    <img alt="pypi" src="https://img.shields.io/pypi/v/poetry-dotenv?logo=pypi">\n  </a>\n  <a href="https://github.com/volopivoshenko/poetry-dotenv/releases">\n    <img alt="release" src="https://img.shields.io/github/v/release/volopivoshenko/poetry-dotenv?logo=github">\n  </a>\n  <a href="https://www.sphinx-doc.org/en/master">\n    <img alt="sphinx" src="https://img.shields.io/badge/made_with-Sphinx-1f425f.svg?logo=readthedocs">\n  </a>\n  <a href="https://numpydoc.readthedocs.io/en/latest/format.html">\n    <img alt="numpydoc" src="https://img.shields.io/badge/docstrings-numpy-1f425f.svg?logo=numpy">\n  </a>\n</p>\n\n<p align="center">\n  <a href="https://github.com/psf/black">\n    <img alt="black" src="https://img.shields.io/badge/code_style-black-black.svg?logo=windowsterminal">\n  </a>\n  <a href="https://pycqa.github.io/isort/index.html">\n    <img alt="isort" src="https://img.shields.io/badge/imports-isort-black.svg?logo=windowsterminal">\n  </a>\n  <a href="https://wemake-python-stylegui.de/en/latest/index.html">\n    <img alt="wemake" src="https://img.shields.io/badge/style-wemake-black.svg?logo=windowsterminal">\n  </a>\n  <a href="https://mypy.readthedocs.io/en/stable/index.html">\n    <img alt="mypy" src="https://img.shields.io/badge/mypy-checked-success.svg?logo=python">\n  </a>\n  <a href="https://github.com/pyupio/safety">\n    <img alt="safety" src="https://img.shields.io/badge/safety-checked-success.svg?logo=windowsterminal">\n  </a>\n  <a href="https://github.com/semantic-release/semantic-release">\n    <img alt="semantic_release" src="https://img.shields.io/badge/semantic_release-angular-e10079?logo=semantic-release">\n  </a>\n</p>\n\n<p align="center">\n  <a href="https://github.com/dependabot">\n    <img alt="dependabot" src="https://img.shields.io/badge/dependabot-enable-success?logo=Dependabot">\n  </a>\n  <a href="https://github.com/volopivoshenko/poetry-dotenv/actions/workflows/integration.yaml">\n    <img alt="integration" src="https://img.shields.io/github/workflow/status/volopivoshenko/poetry-dotenv/CI?label=CI&logo=github">\n  </a>\n  <a href="https://github.com/volopivoshenko/poetry-dotenv/actions/workflows/deployment.yaml">\n    <img alt="deployment" src="https://img.shields.io/github/workflow/status/volopivoshenko/poetry-dotenv/CD?label=CD&logo=github">\n  </a>\n  <a href="https://github.com/volopivoshenko/poetry-dotenv/actions/workflows/codeql.yaml">\n    <img alt="codeql" src="https://img.shields.io/github/workflow/status/volopivoshenko/poetry-dotenv/CodeQL?label=codeQL&logo=github">\n  </a>\n  <a href="https://pypi.org/project/poetry-dotenv">\n    <img alt="wheel" src="https://img.shields.io/pypi/wheel/poetry-dotenv?logo=pypi">\n  </a>\n</p>\n\n<p align="center">\n  <a href="https://codecov.io/gh/volopivoshenko/poetry-dotenv">\n    <img alt="coverage" src="https://img.shields.io/codecov/c/gh/volopivoshenko/poetry-dotenv?logo=codecov&token=yyck08xfTN"/>\n  </a>\n  <a href="https://codeclimate.com/github/volopivoshenko/poetry-dotenv/maintainability">\n    <img alt="codeclimate" src="https://img.shields.io/codeclimate/maintainability/volopivoshenko/poetry-dotenv?logo=codeclimate">\n  </a>\n  <a href="https://pypi.org/project/poetry-dotenv">\n    <img alt="downloads" src="https://img.shields.io/pypi/dm/poetry-dotenv?logo=pypi">\n  </a>\n  <a href="https://github.com/volopivoshenko/poetry-dotenv/">\n    <img alt="stars" src="https://img.shields.io/github/stars/volopivoshenko/poetry-dotenv?logo=github">\n  </a>\n</p>\n\n<p align="center">\n  <a href="https://github.com/volopivoshenko/poetry-dotenv/issues">\n    <img alt="issues" src="https://img.shields.io/github/issues/volopivoshenko/poetry-dotenv?logo=github">\n  </a>\n  <a href="https://github.com/volopivoshenko/poetry-dotenv/issues">\n    <img alt="issues" src="https://img.shields.io/github/issues-closed/volopivoshenko/poetry-dotenv?logo=github">\n  </a>\n  <a href="https://github.com/volopivoshenko/poetry-dotenv/pulls">\n    <img alt="pr" src="https://img.shields.io/github/issues-pr/volopivoshenko/poetry-dotenv?logo=github">\n  </a>\n  <a href="https://github.com/volopivoshenko/poetry-dotenv/pulls">\n    <img alt="pr" src="https://img.shields.io/github/issues-pr-closed/volopivoshenko/poetry-dotenv?logo=github">\n  </a>\n  <a href="https://github.com/volopivoshenko/poetry-dotenv/graphs/contributors">\n    <img alt="contributors" src="https://img.shields.io/github/contributors/volopivoshenko/poetry-dotenv?logo=github">\n  </a>\n  <a href="https://github.com/volopivoshenko/poetry-dotenv/commits/main">\n    <img alt="commit" src="https://img.shields.io/github/last-commit/volopivoshenko/poetry-dotenv?logo=github">\n  </a>\n</p>\n\n<p align="center">\n  <a href="https://www.buymeacoffee.com/volopivoshenko" target="_blank">\n    <img alt="buymeacoffee" src="https://img.shields.io/badge/buy_me_-a_coffee-ff6964?logo=buymeacoffee">\n  </a>\n</p>\n\n- [Overview](#overview)\n  - [Usage](#usage)\n- [Installation](#installation)\n\n# Overview\n\n`poetry-dotenv` - is the plugin that automatically loads environment variables from a dotenv file\ninto the environment before `poetry` commands are run.\n\nThis plugin doesn\'t have any dependencies, but therefore it also supports features that\n`python-dotenv` supports (e.g. templates, interpolating variables using `POSIX` variable expansions etc).\n\n## Usage\n\nBy default plugin will load the `.env` file from the current working directory or "higher directories".\n\nTo prevent ``poetry`` from loading the dotenv file, set the ``POETRY_DONT_LOAD_ENV``\nenvironment variable.\n\nIf your dotenv file is located in a different path or has a different name you may set\nthe ``POETRY_DOTENV_LOCATION`` environment variable.\n\n# Installation\n\n```bash\npoetry self add poetry-dotenv\n```\n',
    'author': 'Volodymyr Pivoshenko',
    'author_email': 'volodymyr.pivoshenko@gmail.com',
    'maintainer': 'Volodymyr Pivoshenko',
    'maintainer_email': 'volodymyr.pivoshenko@gmail.com',
    'url': 'https://github.com/volopivoshenko/poetry-dotenv',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<3.11',
}


setup(**setup_kwargs)
